/* netserver.c
 *
 * Networking functions used only by the servers.
 * Server-client protocol functions. 
 */
#include "sim.h"
#include "server.h"

static int process_con_req(int fd, struct in_addr addr, char *hostname);
static int get_con_req(int fd);
static void freeserverlist(simserver *head);
static int look_for_player(char *name, int fd, simserver *sh, simserver **done);
static void vserver_command(int fd, int com, va_list ap);

/* process_connection
 * Receive and process a connection request from a client or other server.
 */
void process_connection(int serverfd)
{
int fd;			/* fd of new connection */
char hostname[256];	/* hostname of connecting machine */
struct in_addr addr;	/* address of connecting machine */
int req;

fd = new_connection(serverfd,hostname,&addr);

if (!check_version(fd)) {
	close(fd);
	return;
	}

/* get and process requests until done
 */
while((req = process_con_req(fd,addr,hostname)) == 1)
	;
if (req != 2)
	close(fd);
}


/* process_con_req
 * Process one connection request.
 * Returns 1 if this request was processed ok and another should be waited for,
 *    2 if this request was the last one and the connection should be kept open
 *    0 if the connection should be closed.
 * See protocol.html for request format
 */
static int process_con_req(int fd, struct in_addr addr, char *hostname)
{
int req;		/* connection request number */

if ((req = get_con_req(fd)) == -1) {
	return 0;
	}

switch(req) {
case con_newplayer: {
	/* This is a new client joinging the game.
	 * Either find them the server they are on, or create them a new
	 * player on this server
	 */
	char name[256];		/* player name from client */
	shape *sh;

	logmessage("newplayer request received from %s",hostname);
	if (fdgetw(fd,name,256) == NULL) {
		logerror("Error reading name from client");
		return 0;
		}
	/* Does a shape of the same name exist on this server?
	 */
	reset_shape();
	while((sh = next_shape()))
		if (strequ(sh->name,name)) {
			/* They are. Tell the player the name of this server
			 */
			fdprintf(fd,"%s\n",mp->server);
			fdprintf(fd,"%d\n",mp->port);
			return 1;
			}

	/* Recursively query every server for either this player's name
	 * or a list of connecting servers.
	 */
	if (!sfork()) {
		simserver *sh;		/* list of servers already checked */
		shape *nsh;

		set_signals(SIG_DFL);

		/* already done THIS one
		 */
		sh = smalloc(sizeof(simserver));
		sh->addr = strdupe(mp->server);
		sh->ip   = mp->ip;
		sh->port = mp->port;
		sh->next = NULL;

		if (look_for_player(name,fd,mp->shead,&sh) == 0) {
			/* Didn't find the player anywhere.. create them
			 * on this server, in a safe place. 
			 */
			int lfd, sn;
			player *npl;
			person *npr;

			lfd = open_tcp_send_ip(mp->ip,mp->port);
			if (lfd == -1) {
				logerror("Error connecting to self!");
				exit(0);
				}
			if (!send_version(lfd)) {
				exit(0);
				}

			/* Create and transfer a new player
			 */
			logmessage("Creating player %s",name);
			fdprintf(lfd,"%d\n",con_transfer_player);
			if (!send_password(lfd,mp->password)) {
				exit(0);
				}
			npl = create_player();
			strcpy(npl->name,name);
			npl->inout = time(NULL);
			write_player(lfd,0,npl);
			free_player(npl);

			/* Create and transfer a new person for this player
			 */
			fdprintf(lfd,"%d\n",con_transfer_person);
			send_password(lfd,mp->password);
			npr = create_person();
			strcpy(npr->name,name);
			sn = rand()%mp->safenodec;
			npr->pos = lookup_node(mp->safenode[sn]);
			npr->email = smalloc(strlen(name)+strlen(hostname)+2);
			sprintf(npr->email,"%s@%s",name,hostname);
			npr->fullname = strdupe(name);
			write_person(lfd,0,npr);
			free_person(npr);

			/* Send the default avatar for this new player to use.
			 * Note - the shape is located at 0,0,0 for the
			 * receiving server to later relocate.
			 *      - the shape is shaded to glass to indicate
			 * that the controlling player is not connected yet.
			 */
			fdprintf(lfd,"%d\n",con_transfer_shape);
			send_password(lfd,mp->password);
			nsh = deep_copy_shape(lookup_shape("_player"));
			strcpy(nsh->name,name);
			shade_shape(nsh,sft_glass);
			write_shape(lfd,0,nsh,lvl_textured,1);
			free_shape(nsh);
			close(lfd);

			/* Tell the client to look here for their player
			 */
			fdprintf(fd,"%s\n",mp->server);
			fdprintf(fd,"%d\n",mp->port);
			}
		freeserverlist(sh);
		exit(0);
		}

	return 0;
	break;
	}

case con_oldplayer: {
	/* This is a client that knows that their player is on this server.
	 * Link them into the game.
	 * OR, it could be a client with the same name as an NPC on this server.
	 * OR, it could be a client with the same name as an object on server.
	 */
	char name[256];		/* player name from client */
	char pass[256];		/* password from client */
	player *p;		/* struct of this player */
	char ilevel[20];	/* 3d info level requested */
	shape *sh;		/* 3d shape of player */

	logmessage("oldplayer request received from %s",hostname);
	if (fdgets(fd,name,256) == NULL) {
		logerror("Invalid name recieved from client");
		return 0;
		}
	strip_return(name);
	if (fdgets(fd,pass,256) == NULL) {
		logerror("Invalid password recieved from client");
		return 0;
		}
	strip_return(pass);

	reset_player();
	while((p = next_player()))
		if (strequ(p->name,name))
			break;
	if (!p) {
		if (lookup_person(name)) {
			/* An NPC already has this name
			 */
			fdprintf(fd,"already\n");
			logerror("Name clash with NPC %s",name);
			return 0;
			}
		else if (lookup_object(name)) {
			/* An object already has this name
			 */
			fdprintf(fd,"already\n");
			logerror("Name clash with object %s",name);
			return 0;
			}
		else {
			/* Player not found
			 */
			fdprintf(fd,"fail\n");
			logerror("Player %s not found",name);
			return 0;
			}
		}
	if (p->fd != -1) {
		/* Player already being controlled
		 */
		fdprintf(fd,"already\n");
		logerror("Player %s already active",p->name);
		return 0;
		}
	if (p->password && !strequ(p->password,pass)) {
		/* Wrong password sent when one was needed
		 */
		fdprintf(fd,"wrong\n");
		logerror("Wrong password from player %s",p->name);
		return 0;
		}
	fdprintf(fd,"ok\n");

	/* Make player visible if logged out for a long time
	 */
	sh = lookup_shape(p->name);
	if (sh->hidden) {
		player *pp;

		sh->hidden = 0;
		reset_player();
		while((pp = next_online_player()) != NULL) {
			server_command(pp->fd,com_addshape,sh->name);
			write_shape(pp->fd,0,sh,pp->ilevel,0);
			}
		}

	/* Send map details to client
	 */
	write_nodes(fd);
	write_paths(fd);
	write_persons(fd);
	write_objects(fd);

	/* Send info on their player to client
	 */
	write_player(fd,0,p);

	/* Send 3d info
	 */
	if (fdgetw(fd,ilevel,20) == NULL) {
		logerror("Error reading 3d level from client");
		return 0;
		}
	p->ilevel = atoi(ilevel);
	if (p->ilevel == lvl_none) {
		/* don't send anything */
		}
	else if (p->ilevel == lvl_flat || p->ilevel == lvl_textured) {
		write_shapes(fd,p->ilevel,0);
		}
	else {
		logerror("Unknown 3d info level number from client : %s",
			 ilevel);
		return 0;
		}

	p->fd = fd;
	p->ip = addr;
	p->inout = time(NULL);
	logmessage("Player %s has entered the map",p->name);

	/* Return this player's shape to solid form
	 */
	shade_shape(lookup_shape(p->name),sft_flat);
	server_command_all(com_shdshape,p->name,sft_flat);
	return 2;
	break;
	}

case con_locateplayer: {
	/* This is another server looking for a player on this server.
	 * Either tell them that the player is here, or give a list of
	 * other servers to try.
	 */
	char name[256];
	shape *sh;
	simserver *s;

	logmessage("locateplayer request received from %s",hostname);
	if (fdgetw(fd,name,256) == NULL) {
		logerror("Invalid name recieved from server");
		return 0;
		}

	/* Look for a shape of this name
	 */
	reset_shape();
	while((sh = next_shape()))
		if (strequ(sh->name,name)) {
			fdprintf(fd,"found\n");
			return 1;
			}

	/* not found. send a list of connecting servers for the other
	 * server to try
	 */
	fdprintf(fd,"unknown\n");
	for(s=mp->shead; s; s=s->next)
		fdprintf(fd,"%s\n%d\n",s->addr,s->port);
	fdprintf(fd,"done\n0\n");
	break;
	}

case con_transfer_player: {
	/* Another server transferring a player struct
	 */
	player *npl;

	logmessage("transfer player request received from %s",hostname);
	if (!check_password(fd,mp->password)) {
		return 0;
		}
	if (parse_player(fd,&npl) != 0) {
		logerror("Error parsing transferred player");
		return 0;
		}
	add_player(npl);
	break;
	}

case con_transfer_person: {
	/* Another server transferring a person struct
	 */
	person *npr;
	player *pp;

	logmessage("transfer person request received from %s",hostname);
	if (!check_password(fd,mp->password)) {
		return 0;
		}
	if (parse_person(fd,&npr) != 0) {
		logerror("Error parsing transferred person");
		return 0;
		}
	/* Send new person to all other players
	 */
	reset_player();
	while((pp = next_online_player())) {
		server_command(pp->fd,com_addperson,npr->name);
		write_person(pp->fd,0,npr);
		}
	add_person(npr);
	break;
	}

case con_transfer_shape: {
	/* Another server transferring a shape struct. This could be a player's
	 * avatar, or an item in a player's inventory.
	 */
	shape *nsh;
	person *npr;
	player *pp;

	logmessage("transfer shape request received from %s",hostname);
	if (!check_password(fd,mp->password)) {
		return 0;
		}
	if (parse_shape(fd,&nsh) != 0) {
		logerror("Error parsing transferred shape");
		return 0;
		}

	if ((npr = lookup_person(nsh->name))) {
		/* This shape is some player's avatar. Position it at the
		 * person's location.
		 */
		move_shape(nsh,npr->pos->x,npr->pos->y,npr->pos->z);
		}

	if (!nsh->hidden && nsh->name[0] != '_') {
		/* Send new shape to all other players
		 */
		reset_player();
		while((pp = next_online_player())) {
			if (pp->ilevel != lvl_none) {
				server_command(pp->fd,com_addshape,nsh->name);
				write_shape(pp->fd,0,nsh,pp->ilevel,0);
				}
			}
		}
	add_shape(nsh);
	break;
	}

case con_transfer_object: {
	/* Another server transferring an object
	 */
	object *nob;
	player *pp;

	logmessage("transfer object request received from %s",hostname);
	if (!check_password(fd,mp->password)) {
		return 0;
		}
	if (parse_object(fd,&nob) != 0) {
		logerror("Error parsing transferred object");
		return 0;
		}

	/* Send new object to all other players
	 */
	reset_player();
	while((pp = next_online_player())) {
		server_command(pp->fd,com_addobject,nob->name);
		write_object(pp->fd,0,nob);
		}

	add_object(nob);
	break;
	}

case con_transfer_action: {
	/* Another server transferring an object action
	 */
	action *nac;

	logmessage("transfer action request received from %s",hostname); 
	if (!check_password(fd,mp->password)) {
		return 0;
		}
	if (parse_action(fd,&nac) != 0) {
		logerror("Error parsing transferred action");
		return 0;
		}

	add_action(nac);
	break;
	}

case con_transfer_obdata: {
	/* Another server transferring object data
	 */
	char objname[256], datname[256], portstr[256], lenstr[256];
	int port,len;
	char filbuf[256];
	int f;		/* fd for received data */

	logmessage("transfer object data request received from %s",hostname);
	if (!check_password(fd,mp->password)) {
		return 0;
		}
	if (fdgetw(fd,objname,256) == NULL) {
		logerror("Error reading name of object data is for");
		return 0;
		}
	if (fdgetw(fd,datname,256) == NULL) {
		logerror("Error reading data name for object");
		return 0;
		}
	if (fdgetw(fd,portstr,256) == NULL || (port = atoi(portstr)) == 0) {
		logerror("Error reading port for object data transfer");
		return 0;
		}
	if (fdgetw(fd,lenstr,256) == NULL || (len = atoi(lenstr)) == 0) {
		logerror("Error reading length for object data transfer");
		return 0;
		}
	sprintf(filbuf,"%s/object_data/%s/%s",mp->mapdir,objname,datname);
	if ((f = open_data_file(filbuf)) == -1) {
		logerror("Error creating object data file %s",filbuf);
		break;
		}
	if (!sfork()) {
		/* Receive data to file concurrently
		 */
		int rv;

		set_signals(SIG_DFL);
		rv = recv_data(addr,port,f,len);
		close(f);
		if (rv == 0) {
			logerror("Error reading object data from server");
			unlink(filbuf);
			}
		exit(0);
		}
	close(f);
	break;
	}

case con_transfer_pldata: {
	/* Another server transferring player data
	 */
	char planame[256], datname[256], portstr[256], lenstr[256];
	int port,len;
	char filbuf[256];
	int f;

	logmessage("transfer player data request received from %s",hostname);
	if (!check_password(fd,mp->password)) {
		return 0;
		}
	if (fdgetw(fd,planame,256) == NULL) {
		logerror("Error reading name of player data is for");
		return 0;
		}
	if (fdgetw(fd,datname,256) == NULL) {
		logerror("Error reading data name for player");
		return 0;
		}
	if (fdgetw(fd,portstr,256) == NULL || (port = atoi(portstr)) == 0) {
		logerror("Error reading port for player data transfer");
		return 0;
		}
	if (fdgetw(fd,lenstr,256) == NULL || (len = atoi(lenstr)) == 0) {
		logerror("Error reading length for player data transfer");
		return 0;
		}
	sprintf(filbuf,"%s/player_data/%s/%s",mp->mapdir,planame,datname);
	if ((f = open_data_file(filbuf)) == -1) {
		logerror("Error creating player data file %s",filbuf);
		break;
		}
	if (!sfork()) {
		/* Receive data to file concurrently
		 */
		int rv;

		set_signals(SIG_DFL);
		rv = recv_data(addr,port,f,len);
		close(f);
		if (rv == 0) {
			logerror("Error reading player data from server");
			unlink(filbuf);
			}
		exit(0);
		}
	close(f);
	break;
	}

case con_shutdown: {
	/* Someone requesting that this server shut down
	 */
	logmessage("shutdown server request received from %s",hostname);
	if (!check_password(fd,mp->password)) {
		return 0;
		}
	kill(getpid(),SIGINT);		/* this will call crash_stop */
	break;
	}

case con_noop: {
	/* Do-nothing request
	 */
	return 1;
	break;
	}

case con_status: {
	/* Request for server status
	 */
	person *pr;

	logmessage("status request received from %s",hostname);
	if (!check_password(fd,mp->password)) {
		return 0;
		}

	/* send person one per line, formatted as
	 * name position fd time ipnum ilevel	for online players
	 * name position 'offline' time		for offline players
	 * name position 'npc'			for npcs
	 */
	reset_person();
	while((pr = next_person())) {
		player *pp = lookup_player(pr->name);

		fdprintf(fd,"\"%s\" \"%s\" ",pr->name,pr->pos->name);
		if (pp) {
			if (pp->fd != -1)
				fdprintf(fd,"%d %ld %lu %d\n",pp->fd,pp->inout,
					 pp->ip.s_addr,pp->ilevel);
			else
				fdprintf(fd,"offline %ld\n",pp->inout);
			}
		else
			fdprintf(fd,"npc\n");
		}
	fdputc(fd,'\n');

	return 1;
	break;
	}

case con_getshapes: {
	/* Request for 3d shapes
	 */
	char ilevelstr[20];
	int ilevel;

	logmessage("getshapes request received from %s",hostname);

	if (fdgetw(fd,ilevelstr,20) == NULL) {
		logerror("Error reading 3d level for getshapes");
		return 0;
		}
	ilevel = atoi(ilevelstr);
	if (ilevel == lvl_none) {
		/* Don't send anything!? */
		}
	else if (ilevel == lvl_flat || ilevel == lvl_textured) {
		write_shapes(fd,ilevel,0);
		}
	else {
		logerror("Unknown 3d info level : %s",ilevelstr);
		return 0;
		}
	return 1;
	break;
	}

default: {
	logerror("Unknown request number : %d",req);
	return 0;
	break;
	}

	} /* end switch */
return 1;
}


/* get_con_req
 * Read connection request number from client. Returns -1 on error.
 */
static int get_con_req(int fd)
{
char req[20];

if (fdgetw(fd,req,20) == NULL) {
	/* logerror("Error reading connection request number"); */
	return -1;
	}
return atoi(req);
}


/* look_for_player
 * Recursively query a list of servers, retreiving from each either
 * a "found", or an "unknown" followed by list of other servers to try.
 * name - player to look for
 * fd   - connection to client, to report location of their player on
 *        if found
 * sh   - head of a list of servers to try
 * done - head of a list of servers already tried
 */ 
static int look_for_player(char *name, int fd, simserver *sh, simserver **done)
{
for(; sh; sh=sh->next) {
	simserver *dp;		/* for traversing done list */
	simserver *ns;		/* new server to add to done list */
	int sfd;		/* fd for connecting to other server */
	char res[20];		/* response on player from other server */

	/* Is this server on the done list?
	 */
	for(dp=*done; dp; dp=dp->next)
		if (dp->ip.s_addr == sh->ip.s_addr && dp->port == sh->port)
			break;	/* it is */
	if (dp)
		continue;	/* already done this one */

	/* Add this server to the done list
	 */
	ns = scalloc(1,sizeof(simserver));
	ns->addr = strdupe(sh->addr);
	ns->ip = sh->ip;
	ns->port = sh->port;
	ns->next = *done;
	*done = ns;

	/* Try to establish connection to other server.
	 */
	logmessage("Looking on %s port %d",sh->addr,sh->port);
	sfd = open_tcp_send_ip(sh->ip,sh->port);
	if (sfd == -1) {
		logerror("Couldn't contact server %s for lookup\n",sh->addr);
		fdprintf(fd,"fail\n");
		fdprintf(fd,"%s\n",sh->addr);
		return -1;
		}

	/* Check version on remote server
	 */
	if (!send_version(sfd)) {
		close(sfd);
		fdprintf(fd,"fail\n");
		fdprintf(fd,"%s\n",sh->addr);
		return -1;
		}

	/* Ask the remote server if it knows about the player
	 */
	fdprintf(sfd,"%d\n",con_locateplayer);
	fdprintf(sfd,"%s\n",name);
	if (fdgetw(sfd,res,20) == NULL) {
		logerror("Error reading locateplayer response from server %s",
			 sh->addr);
		close(sfd);
		fdprintf(fd,"fail\n");
		fdprintf(fd,"%s\n",sh->addr);
		return -1;
		}
	if (strequ(res,"found")) {
		/* YES! found the player. Tell the client the location of
		 * the server they are on.
		 */
		fdprintf(fd,"%s\n",sh->addr);
		fdprintf(fd,"%d\n",sh->port);
		return 1;
		}
	else {
		/* Read a list of other possible servers to try
		 */
		simserver *ps = NULL;	/* list of possible servers */
		char paddr[256];
		char pport[256];

		while(1) {
			simserver *nps;
			struct hostent *hent;

			if (fdgetw(sfd,paddr,256) == NULL)
				break;
			if (fdgetw(sfd,pport,256) == NULL)
				break;
			if (strequ(paddr,"done"))
				break;
			nps = scalloc(1,sizeof(simserver));
			nps->addr = strdupe(paddr);
			nps->port = atoi(pport);
			if ((hent = gethostbyname(paddr)) == NULL) {
				logerror("Couldn't get ip number of %s for "
					 "lookup",paddr);
				return 0;
				}
			nps->ip = *(struct in_addr *)hent->h_addr;
			nps->next = ps;
			ps = nps;
			}
		close(sfd);

		if (look_for_player(name,fd,ps,done))
				return 1;

		/* Free list of possible servers
		 */
		freeserverlist(ps);
		}
	}
return 0;
}


/* freeserverlist
 * Recursively free a list of servers
 */
static void freeserverlist(simserver *head)
{
if (head) {
	freeserverlist(head->next);
	if (head->addr)
		free(head->addr);
	free(head);
	}
}


/* server_command
 * Send a command line to a client, using the format described in protocol.html
 */
void server_command(int fd, int com, ...)
{
va_list ap;

va_start(ap,com);
vserver_command(fd,com,ap);
va_end(ap);
}


/* server_command_all
 * Send a command to all active players
 */
void server_command_all(int com, ...)
{
va_list ap;
player *ph;

reset_player();
while((ph = next_online_player())) {
	va_start(ap,com);
	vserver_command(ph->fd,com,ap);
	va_end(ap);
	}
}



/* vserver_command
 * Does the real work of sending a command to a client. Accepted commands are
 * com_move       : char *node
 * com_msg        : char **msgv, int msgc
 * com_err        : int error_num
 * com_addobject  : char *object
 * com_addperson  : char *person
 * com_addshape   : char *shape
 * com_moveobject : char *object, char *node
 * com_moveperson : char *person, char *node
 * com_moveshape  : char *shape, double dx, double dy, double dz
 * com_delobject  : char *object
 * com_delperson  : char *person
 * com_delshape   : char *shape
 * com_newserv    : char *host, int port
 * com_getobject  : char *object, char *person
 * com_dropobject : char *object, char *person
 * com_rotshape   : char *shape, double ang, double x, double y, double z
 * com_giveobject : char *object, char *giver, char *receiver
 * com_displaymime: char *type, int size, int port
 * com_text       : int linec
 * com_actlist    : char *object, char **actc, int actc
 * com_colshape   : char *shape, int r, int g, int b
 * com_shdshape   : char *name, int shade
 * com_ls         : char **datv, int datc
 * com_chpath     : char *node1, char *node2, int closed
 * com_shutdown   : 
 * com_email      : char *person, char *address
 * com_fullname   : char *person, char *name
 * com_teleport   : char *node
 */
static void vserver_command(int fd, int com, va_list ap)
{
static mesgargs cma[] ={
	{com_move,{arg_str,0}},
	{com_msg,{arg_strs,arg_int,0}},
	{com_err,{arg_int,0}},
	{com_addobject,{arg_str,0}},
	{com_addperson,{arg_str,0}},
	{com_addshape,{arg_str,0}},
	{com_moveobject,{arg_str,arg_str,0}},
	{com_moveperson,{arg_str,arg_str,0}},
	{com_moveshape,{arg_str,arg_dbl,arg_dbl,arg_dbl,0}},
	{com_delobject,{arg_str,0}},
	{com_delperson,{arg_str,0}},
	{com_delshape,{arg_str,0}},
	{com_newserv,{arg_str,arg_int,0}},
	{com_getobject,{arg_str,arg_str,0}},
	{com_dropobject,{arg_str,arg_str,0}},
	{com_rotshape,{arg_str,arg_dbl,arg_dbl,arg_dbl,arg_dbl,0}},
	{com_giveobject,{arg_str,arg_str,arg_str,0}},
	{com_displaymime,{arg_str,arg_int,arg_int,0}},
	{com_text,{arg_int,0}},
	{com_actlist,{arg_str,arg_strs,arg_int,0}},
	{com_colshape,{arg_str,arg_int,arg_int,arg_int,0}},
	{com_shdshape,{arg_str,arg_int,0}},
	{com_ls,{arg_strs,arg_int,0}},
	{com_chpath,{arg_str,arg_str,arg_int,0}},
	{com_shutdown,{0}},
	{com_email,{arg_str,arg_str,0}},
	{com_fullname,{arg_str,arg_str,0}},
	{com_teleport,{arg_str,0}},
	};

send_message(fd,com,ap,cma,sizeof(cma)/sizeof(mesgargs));
}


/* server_message
 * Convenience function for sending a formatted string to a player
 */
void server_message(int fd, char *fmt, ...)
{
va_list ap;
char buf[1024];
char **word;
int wordc;

va_start(ap,fmt);
vsprintf(buf,fmt,ap);
va_end(ap);
parse_string(buf,&word,&wordc);
server_command(fd,com_msg,word,wordc);
}


/* store_to_end
 * Read lines of data to a temp file, stopping when a line consisting only
 * of 'end' is read. Returns the name of the temp file.
 */
char *store_to_end(int fd)
{
static char tname[256];
int tfd;

tmpnam(tname);
tfd = creat(tname,0700);
while(1) {
	char line[1024];

	fdgets(fd,line,1024);
	strip_return(line);
	if (strequ(line,"end")) break;
	fdprintf(tfd,"%s\n",line);
	}
return tname;
}

